import gymnasium as gym
from gymnasium import spaces
import numpy as np
class Moral_HazardEnv(gym.Env):
    metadata = {"render.modes": ["human"]}
    # low 参数表示该维度的最低值，high 参数表示该维度的最高值。每个维度都可以有不同的含义，例如：
    # 第1维 - 私人收益
    # 第2维 - 项目成功概率
    # action_space :
    #0 - 不努力
    #1 - 努力
    def __init__(self):
        self.observation_space = spaces.Box(low=np.array([0, 0]), high=np.array([1, 1]), dtype=np.float32, shape=(2,))
        self.action_space = spaces.Discrete(2)
        self.additional_info = None
    def transition(self):
        pass
    def reward(self, x, a, y):
        pass
    def step(self, action):
        if action == 0: # 努力
            private_profit = np.clip(np.random.normal(0.5, 0.1),0.01,0.99)# 私人收益在 0.5 左右波动
            project_success =  np.clip(np.random.normal(0.8, 0.1),0.01,0.99) # 项目成功概率在 0.8 左右波动
        else: # 卸责
            private_profit = np.clip(np.random.normal(0.7, 0.1),0.01,0.99)# 私人收益在 0.7 左右波动
            project_success = np.clip(np.random.normal(0.6, 0.1),0.01,0.99) # 项目成功概率在 0.6 左右波动
        reward = private_profit + project_success # 计算总回报
        done = False # 环境不结束
        info = {}
        truncated = False
        terminated = False
        return np.array([private_profit, project_success]),reward, terminated, truncated, info
    def reset(self, seed=None, options=None):
        super().reset(seed=seed)
        observation = self.observation_space.sample()
        info = {}
        #print(observation)
        return observation,info  # reward, done, info can't be included
    def render(self, mode="human"):
        pass
    def close(self):
        pass
if __name__ == '__main__':
    a=Moral_HazardEnv()
    print(a.observation_space.sample())
    n_observations = a.observation_space.shape[0]
    n_actions = a.action_space.n    
    print(a.action_space)


import gymnasium as gym
import math
import random
import matplotlib
import matplotlib.pyplot as plt
from collections import namedtuple, deque
from itertools import count
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from cpagym.envs.asymmetricInfo_env import *

class DQN(nn.Module):
    def __init__(self, n_observations, n_actions):
        super(DQN, self).__init__()
        self.layer1 = nn.Linear(n_observations, 128)
        self.layer2 = nn.Linear(128, 128)
        self.layer3 = nn.Linear(128, n_actions)

    # Called with either one element to determine next action, or a batch
    # during optimization. Returns tensor([[left0exp,right0exp]...]).
    def forward(self, x):
        x = F.relu(self.layer1(x))
        x = F.relu(self.layer2(x))
        return self.layer3(x)
    
Transition = namedtuple('Transition',
                        ('state', 'action', 'next_state', 'reward'))
class ReplayMemory(object):

    def __init__(self, capacity):
        self.memory = deque([], maxlen=capacity)
    def push(self, *args):
        """Save a transition"""
        self.memory.append(Transition(*args))
    def sample(self, batch_size):
        return random.sample(self.memory, batch_size)
    def __len__(self):
        return len(self.memory)
def main():
    print("hi")

if __name__ == '__main__':
    main()


from cpagym.train.DQN import *
from cpagym.envs.asymmetricInfo_env import *
class Agent:
    """
    Base Agent class handeling the interaction with the environment
    Args:
        env: training environment
        replay_memory: replay buffer storing experiences
    """
    def __init__(self, env: gym.Env, replay_memory: ReplayMemory) -> None:
        self.env = env
        self.replay_memory = replay_memory
        self.reset()
        self.state = self.env.reset()
    def reset(self) -> None:
        """ Resents the environment and updates the state"""
        self.state = self.env.reset()
    def get_action(self, net: nn.Module, epsilon: float, device: str) -> int:
        """
        Using the given network, decide what action to carry out
        using an epsilon-greedy policy
        Args:
            net: DQN network
            epsilon: value to determine likelihood of taking a random action
            device: current device
        Returns:
            action
        """
        if np.random.random() < epsilon:
            action = self.env.action_space.sample()
        else:
            state = torch.tensor([self.state])

            if device not in ['cpu']:
                state = state.cuda(device)

            q_values = net(state)
            _, action = torch.max(q_values, dim=1)
            action = int(action.item())
        return action
    @torch.no_grad()
    def play_step(self, net: nn.Module, epsilon: float = 0.0, device: str = 'cpu'):
        """
        Carries out a single interaction step between the agent and the environment
        Args:
            net: DQN network
            epsilon: value to determine likelihood of taking a random action
            device: current device
        Returns:
            reward, done
        """

        action = self.get_action(net, epsilon, device)

        # do step in the environment
        new_state, reward, done, _ = self.env.step(action)

        exp = Experience(self.state, action, reward, done, new_state)

        self.replay_memory.append(exp)

        self.state = new_state
        if done:
            self.reset()
        return reward, done